// neural_network.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"  
#include <time.h>
#include <stdio.h>

#define _USE_MATH_DEFINES
#include <cmath>

#include <conio.h>
#include <windows.h>


class unit {


public:
	unit();
	~unit();
	int unit_number;
	void build_wmatrix();			
	void move();					
	void backpropagation(double*, double);
	void save_wmatrix_in_genome();
	void print_genome_to_file();

	double* genome;
	double* input;					
	double* hid_layer1_sum;

	double* hid_layer1;

	double* bias_hid_layer1;   

	double* bias_output;		
	double* output_sum;		
	double* output;				
	double** wmatrix1;			

	double** wmatrix2;			

};


//global variables

int n_neur_input=2;
int n_neur_hid_layer1;
int n_neur_output=2;
int  nval_gen_unit =(n_neur_input+1)*n_neur_hid_layer1+(n_neur_hid_layer1+1)*n_neur_output;




//unit class


unit::unit() {
	genome = new double[nval_gen_unit];

	input = new double[n_neur_input+1] ;					

	hid_layer1_sum=new double[n_neur_hid_layer1+1];

	hid_layer1=new double[n_neur_hid_layer1+1];

	output_sum=new double[n_neur_output];		
	output = new double[n_neur_output] ;				

	wmatrix1= new double*[n_neur_hid_layer1];
	for(int i=0;i<n_neur_hid_layer1; i++) wmatrix1[i]=new double[n_neur_input+1];

	wmatrix2= new double*[n_neur_output];
	for(int i=0;i<n_neur_output; i++) wmatrix2[i]=new double[n_neur_hid_layer1+1];

}

void unit:: build_wmatrix() {


	int index=0;


	//construction of wmatrix1

	for(int i=0;i<n_neur_hid_layer1;i++){

		for(int j=0;j<n_neur_input+1;j++){

			wmatrix1[i][j]=genome[index];
			index++;
		}

	}




	//construction of wmatrix2  

	for(int i=0;i<n_neur_output;i++){

		for(int j=0;j<n_neur_hid_layer1+1;j++){

			wmatrix2[i][j]=genome[index];

			index++;
		}

	}


	if(index!=nval_gen_unit) {   // just to avoid errors 
		printf("Error in program source - build_wmatrix routine or nval_gen_unit setting");
		exit(0);	
	}



	//bias initialization (this is important because bias units must be 1 forever; it is comfortable to treat bias as a weighted connection to a unit that gives 1 forever)

	input[n_neur_input]=1;
	hid_layer1[n_neur_hid_layer1]=1;

}

void unit::move(){



	for(int i=0;i<n_neur_hid_layer1;i++){									
		double sum=0;
		for(int j=0;j<n_neur_input+1;j++) sum+=wmatrix1[i][j]*input[j];

		hid_layer1_sum[i]=sum;
	}

	//transfer function hid_layer1 neurons (sigmoid)

	for(int i=0;i<n_neur_hid_layer1;i++){	
		hid_layer1[i]=1/(1+exp(-(hid_layer1_sum[i])))   ;
	}


	//matrix product wmatrix2 * hid_layer1;
	for(int i=0;i<n_neur_output;i++){									
		double sum=0;
		for(int j=0;j<n_neur_hid_layer1+1;j++) sum+=wmatrix2[i][j]*hid_layer1[j];
		output_sum[i]=sum;
	}


	//transfer function output neurons (sigmoid)

	for(int i=0;i<n_neur_output;i++){	
		output[i]=1/(1+exp(-(output_sum[i]))) ;  
	}


}



void unit::backpropagation(double *example_output, double learning_coefficient){

	double* Err_H=new double[n_neur_hid_layer1+1];

	for(int j=0;j<n_neur_hid_layer1+1;j++) Err_H[j]=0;  // should this instruction be here?

	for(int j=0;j<n_neur_output; j++){

		double delta= (example_output[j]-output[j])*output[j]*(1-output[j]);
		for(int k=0;k<n_neur_hid_layer1+1;k++){
			Err_H[k]=Err_H[k]+(delta*wmatrix2[j][k]); //check matrix indices!!!
			wmatrix2[j][k]=wmatrix2[j][k]+(learning_coefficient*delta*hid_layer1[k]);
		}


	}



	for(int k=0;k<n_neur_hid_layer1;k++){

		double delta= Err_H[k]*hid_layer1[k]*(1-hid_layer1[k]);

		for(int i=0;i<n_neur_input+1;i++){
			wmatrix1[k][i]=wmatrix1[k][i]+(learning_coefficient*delta*input[i]);//check matrix indices!!!
		}



	}

	delete[] Err_H;



}





void unit::save_wmatrix_in_genome(){

	int index=0;


	//wmatrix1

	for(int i=0;i<n_neur_hid_layer1;i++){

		for(int j=0;j<n_neur_input+1;j++){

			genome[index]=wmatrix1[i][j];
			index++;
		}

	}




	//wmatrix2

	for(int i=0;i<n_neur_output;i++){

		for(int j=0;j<n_neur_hid_layer1+1;j++){

			genome[index]=wmatrix2[i][j];

			index++;
		}

	}


	if(index!=nval_gen_unit) {
		printf("Error in program source - build_wmatrix routine or nval_gen_unit setting");
		exit(0);	
	}



	printf("\n\nNEURAL NET STORED IN GENOME\n");
}


void unit::print_genome_to_file(){

	FILE* genome_file;

	genome_file= fopen("neural_net_genome.txt","w");

	fprintf(genome_file,"%i\n",n_neur_input);	
	fprintf(genome_file,"%i\n",n_neur_hid_layer1);	
	fprintf(genome_file,"%i\n",n_neur_output);	
	fprintf(genome_file,"%i\n",nval_gen_unit);

	for(int i=0;i<nval_gen_unit;i++) fprintf(genome_file,"%f\n",genome[i]);

	fprintf(genome_file,"End of file");

	fclose(genome_file);
	printf("\n\nFILE PRINTED\n");

}

unit:: ~unit(){
	delete[] genome;
	delete[] input;					
	delete[] hid_layer1_sum;

	delete[] hid_layer1;

	delete[] output_sum;		
	delete[] output;				

	for(int i = 0; i < n_neur_hid_layer1; ++i)  delete [] wmatrix1[i];
	delete [] wmatrix1;


	for(int i = 0; i < n_neur_output; ++i)  delete [] wmatrix2[i];
	delete [] wmatrix2;

}



double rand_in_interval(double a, double b){

	double random_value = a + (rand()/ ((double)RAND_MAX))*(b-a);

	return random_value;
}


void x_and_time_calc(double speed,double angle,double* x,double* time){
	double g= 9.81;
	if(x!=NULL)		*x= speed*speed*sin(2*angle) /g;  
	if(time!=NULL)	*time=2*speed*sin(angle)/g;
}

double reward_calculation(double speed, double angle,double target_x,double target_time){    

	if(speed<0 || angle<0 || angle>M_PI/2) return 0;

	//x and time calculation
	double x;
	double time;
	double x_error_width=10;
	double time_error_width=2;

	x_and_time_calc(speed,angle,&x,&time);

	double reward = 1- (fabs(x - target_x) / x_error_width) -(fabs(time - target_time) / time_error_width); 
	if (reward<0) reward=0;

	return reward;
}



int _tmain(int argc, _TCHAR* argv[])
{



	int n_examples;
	int n_epochs;
	double target_x_min=0;
	double target_x_max=1020.5;    //max distance used in examples
	double target_time_min=0;
	double target_time_max=20.40;  //max flight time used in examples
	double speed_max=100;
	double angle_max=M_PI/2;



	//random engine initialization 
	int seed= time(NULL);
	srand(seed);		
	for (int i=0; i<(seed%10000); i++) rand();  
	printf("first random extracted: %i",rand());  
	printf("\nrandom generator seed=%i\n",seed);



	printf("\nNumber of units in hidden layer:\n");
	scanf("%i",&n_neur_hid_layer1);
	nval_gen_unit=(n_neur_input+1)*n_neur_hid_layer1+(n_neur_hid_layer1+1)*n_neur_output;

	printf("Number of random examples to train the neural net:\n");
	scanf("%i",&n_examples);

	double reward_threshold;
	printf("Requested average reward threshold:\n");
	scanf("%lf",&reward_threshold);



	unit* unit1=new unit;
	double speed;
	double angle;

	double** examples_output=new double*[n_examples];
	for(int i=0; i<n_examples; i++) examples_output[i]=new double[n_neur_output];

	double** examples_input= new double*[n_examples];
	for(int i=0;i<n_examples;i++) examples_input[i]=new double[n_neur_input];




	//initialization of weight matrix (this is to prevent to start with all weights zero or with absurd values)
	for(int i=0;i<nval_gen_unit;i++) unit1->genome[i]=rand_in_interval(-0.5,0.5);
	unit1->build_wmatrix();




	for(int i=0; i<n_examples;i++){

		//plausible example generation

		double temp_speed=rand_in_interval(0,speed_max);
		double temp_angle=rand_in_interval(0,angle_max);

		double target_x;
		double target_time;

		x_and_time_calc(temp_speed,temp_angle,&target_x,&target_time);

		unit1->input[0]= target_x/target_x_max;
		unit1->input[1]= target_time/target_time_max;

		examples_input[i][0]= target_x/target_x_max;
		examples_input[i][1]= target_time/target_time_max;

		examples_output[i][0]=temp_speed/speed_max;
		examples_output[i][1]=temp_angle/angle_max;



	}




	//backpropagation



	printf("\nStarting backpropagation\n");



	double average_reward=0;

	int epochs_count=0;

	do {

		double reward_sum=0;

		if(epochs_count%1000==0) printf("\nEpoch number %i",epochs_count);

		for(int example_count=0;example_count<n_examples;example_count++) {

			unit1->input[0]= examples_input[example_count][0];	
			unit1->input[1]= examples_input[example_count][1];	

			unit1->move();
			reward_sum+=reward_calculation(unit1->output[0]*speed_max,unit1->output[1]*angle_max,unit1->input[0]*target_x_max,unit1->input[1]*target_time_max);

			unit1->backpropagation(examples_output[example_count],0.3);

			if(epochs_count%1000==0) { //just to see what the program is doing without printing too much stuff on the screen
				if(example_count%10==0) printf("\nexample %i\ttarget x=%f\ttarget time=%f\treward=%f",example_count,unit1->input[0]*target_x_max,unit1->input[1]*target_time_max,reward_calculation(unit1->output[0]*speed_max,unit1->output[1]*angle_max,unit1->input[0]*target_x_max,unit1->input[1]*target_time_max));

			}

		}

		if(epochs_count%1000==0) printf("\nAverage reward=%f\n\n",average_reward);

		average_reward=reward_sum/n_examples;

		epochs_count++;


	}while(average_reward<reward_threshold);


	printf("\nreward threshold=%f\tepochs needed=%i\n",reward_threshold,epochs_count);


	if(kbhit()) getch();  

	char c;

	do{

		printf("\n\nPress 's' to save neural net to file neural_net_genome.txt, 't' to test neural net, 'e' to close program\n");


		do{
		}while(!kbhit());

		c=getch();

		if(c=='s'){

			unit1->save_wmatrix_in_genome();
			unit1->print_genome_to_file();

		} else if(c=='t'){

			for(int k=1;k<=10;k++){

				double temp_speed=rand_in_interval(0,speed_max);
				double temp_angle=rand_in_interval(0,angle_max);

				double target_x;
				double target_time;

				x_and_time_calc(temp_speed,temp_angle,&target_x,&target_time);

				printf("\nTest number %i:\n",k);
				printf("expected speed= %f\nexpected angle= %f\nexpected x= %f\nexpected time= %f\n",temp_speed,temp_angle,target_x,target_time);

				unit1->input[0]=target_x/target_x_max;
				unit1->input[1]=target_time/target_time_max;

				unit1->move();

				double unit1_speed= unit1->output[0]*speed_max;
				double unit1_angle= unit1->output[1]*angle_max;

				double unit1_time;
				double unit1_x;

				x_and_time_calc(unit1_speed,unit1_angle,&unit1_x,&unit1_time);

				double reward= reward_calculation(unit1_speed,unit1_angle,target_x,target_time);

				printf("unit1 speed= %f\nunit1 angle= %f\nunit1 x= %f\nunit1 time= %f\nreward= %f\n",unit1_speed,unit1_angle,unit1_x,unit1_time,reward);

			}




		}





	}while(c!='e');


	//delete commands


	for(int i=0; i<n_examples; i++) delete[] examples_output[i];  // to be checked
	delete[] examples_output;

	for(int i=0; i<n_examples; i++) delete[] examples_input[i];  // to be checked
	delete[] examples_input;



	return 0;
}




